#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include "tb.h"

#define TEST_ROWS	20000

int main()
{
  DWORD dw=GetTickCount();

	struct tb_table t, t1;
	struct tb_column col[5];
	int i;
	time_t tt1, tt2;
	double tott;

	tott = 0.0;

	srand(time(NULL));

	tb_initialize(&t);
	tb_initialize(&t1);
	
	printf("test 0: discard immediately after initialize\n");
	tb_discard(&t);
	tb_discard(&t1);
	
	strcpy(col[0].name, "INT");
	col[0].type = tb_type_int;
	strcpy(col[1].name, "DBL");
	col[1].type = tb_type_double;
	strcpy(col[2].name, "CHR");
	col[2].type = tb_type_char;
	col[2].size = 128;

	tb_describe(&t, col, 3, 0);

	printf("INT col number %d\n", tb_findcol(&t, "INT"));
	printf("DBL col number %d\n", tb_findcol(&t, "DBL"));
	printf("CHR col number %d\n", tb_findcol(&t, "CHR"));
	printf("HR col number %d\n", tb_findcol(&t, "HR"));
	printf("\n");

	printf("test 1: insert %d rows to the begining\n", TEST_ROWS);
	time(&tt1);
	for(i = 0; i < TEST_ROWS; i++)
	{
		tb_insert(&t, 0);
	}
	time(&tt2);
	printf("%d rows inserted\n", tb_getnrows(&t));
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	printf("test 2: insert %d rows to the end\n", TEST_ROWS);
	time(&tt1);
	for(i = 0; i < TEST_ROWS; i++)
	{
		tb_append(&t);
	}
	time(&tt2);
	printf("%d rows inserted\n", tb_getnrows(&t));
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	printf("test 3: remove all rows from the begining\n");
	time(&tt1);
	while(tb_remove(&t, 0) >= 0)
		;
	time(&tt2);
	printf("%d rows in the table\n", tb_getnrows(&t));
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	printf("test 4: insert %d rows to the end\n", TEST_ROWS);
	time(&tt1);
	for(i = 0; i < TEST_ROWS; i++)
	{
		tb_append(&t);
	}
	time(&tt2);
	printf("%d rows inserted\n", tb_getnrows(&t));
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	printf("test 3: insert %d rows to the begining\n", TEST_ROWS);
	time(&tt1);
	for(i = 0; i < TEST_ROWS; i++)
	{
		tb_insert(&t, 0);
	}
	time(&tt2);
	printf("%d rows inserted\n", tb_getnrows(&t));
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	printf("test 5: remove all rows from the end\n");
	time(&tt1);
	while(tb_getnrows(&t) > 0)
		tb_remove(&t, tb_getnrows(&t) - 1);
	time(&tt2);
	printf("%d rows in the table\n", tb_getnrows(&t));
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	printf("test 6: insert %d rows randomly\n", TEST_ROWS);
	time(&tt1);
	for(i = 0; i < TEST_ROWS; i++)
	{
		tb_insert(&t, rand()%(i + 1));
	}
	time(&tt2);
	printf("%d rows inserted\n", tb_getnrows(&t));
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	printf("test 7: remove %d rows randomly\n", TEST_ROWS);
	time(&tt1);
	while(tb_getnrows(&t) > 0)
		tb_remove(&t, rand()%tb_getnrows(&t));
	time(&tt2);
	printf("%d rows in the table\n", tb_getnrows(&t));
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	printf("test 8: tb_clear vs tb_remove\n");
	for(i = 0; i < TEST_ROWS; i++)
		tb_append(&t);
	time(&tt1);
	tb_clear(&t);
	time(&tt2);
	printf("%d rows in the table\n", tb_getnrows(&t));
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	for(i = 0; i < TEST_ROWS*2; i++)
	{
		char buf[20], *b;

		tb_append(&t);
		tb_setint(&t, i, 0, i);
		tb_setdouble(&t, i, 1, i);
		sprintf(buf, "%x", i);
		b = tb_setchar(&t, i, 2, buf);
		if(strcmp(b, buf))
		{
		    printf("%s != %s\n", b, buf);
		    return 2001;
		}
	}

	printf("test 9: forward scanning (%d rows)\n", tb_getnrows(&t));
	time(&tt1);
	for(i = 0; i < tb_getnrows(&t); i++)
	{
		char buf[20], buf1[20];

		if(i != tb_getint(&t, i, 0))
		{
			printf("i != tb_getint(&t, i, 0) ~ %d != %d\n", i, tb_getint(&t, i, 0));
			return 2001;
		}
		if((double)i != tb_getdouble(&t, i, 1))
		{
			printf("(double)i != tb_getdouble(&t, i, 1) ~ %f != %f\n", (double)i, tb_getdouble(&t, i, 1));
			return 2001;
		}

		sprintf(buf1, "%x", i);
		tb_getchar(&t, i, 2, buf, sizeof(buf));

		if(strcmp(buf, buf1))
		{
			printf("line %d (%x), strcmp(buf, buf1) ~ \"%s\" != \"%s\"\n", i, i, buf, buf1);
			return 2001;
		}
	}
	time(&tt2);
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	printf("test 10: backward scanning (%d rows)\n", tb_getnrows(&t));
	time(&tt1);
	for(i = tb_getnrows(&t) - 1; i >= 0; i--)
	{
		char buf[20], buf1[20];

		if(i != tb_getint(&t, i, 0))
		{
			printf("i != tb_getint(&t, i, 0) ~ %d != %d\n", i, tb_getint(&t, i, 0));
			return 2001;
		}
		if((double)i != tb_getdouble(&t, i, 1))
		{
			printf("(double)i != tb_getdouble(&t, i, 1) ~ %f != %f\n", (double)i, tb_getdouble(&t, i, 1));
			return 2001;
		}

		sprintf(buf1, "%x", i);
		tb_getchar(&t, i, 2, buf, sizeof(buf));

		if(strcmp(buf, buf1))
		{
			printf("strcmp(buf, buf1) ~ \"%s\" != \"%s\"\n", buf, buf1);
			return 2001;
		}
	}
	time(&tt2);
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");

	printf("test 11: random scanning (%d rows)\n", tb_getnrows(&t));
	time(&tt1);
	for(i = 0; i < tb_getnrows(&t); i++)
	{
		char buf[20], buf1[20];
		int k;

		k = rand()%tb_getnrows(&t);

		if(k != tb_getint(&t, k, 0))
		{
			printf("k != tb_getint(&t, k, 0) ~ %d != %d\n", k, tb_getint(&t, k, 0));
			return 2001;
		}
		if((double)k != tb_getdouble(&t, k, 1))
		{
			printf("(double)k != tb_getdouble(&t, k, 1) ~ %f != %f\n", (double)k, tb_getdouble(&t, k, 1));
			return 2001;
		}

		sprintf(buf1, "%x", k);
		tb_getchar(&t, k, 2, buf, sizeof(buf));

		if(strcmp(buf, buf1))
		{
			printf("strcmp(buf, buf1) ~ %s != %s\n", buf, buf1);
			return 2001;
		}
	}
	time(&tt2);
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");
	
	printf("test 12: tb_match\n");
	time(&tt1);
	{
	    struct tb_column col[4];
	    int pos[4], j;
	    
	    strcpy(col[0].name, "CHR");
	    col[0].type = tb_type_char;
	    col[0].size = 128;
	    strcpy(col[1].name, "INT");
	    col[1].type = tb_type_int;
	    strcpy(col[2].name, "DBL");
	    col[2].type = tb_type_double;
	    strcpy(col[3].name, "ANY");
	    col[3].type = tb_type_int;
	    
	    printf("not exact match: %d\n", tb_match(&t, col, 3, pos, 0));
	    for(j = 0; j < 4; j++)
		printf("%d ", pos[j]);
	    printf("\n");
	    printf("exact match (1): %d\n", tb_match(&t, col, 4, pos, 1));
	    printf("exact match (2): %d\n", tb_match(&t, col, 3, pos, 1));
	    
	    col[1].type = tb_type_double;
	    
	    printf("exact match (3): %d\n", tb_match(&t, col, 3, pos, 1));
	}
	time(&tt2);
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");
	
	printf("test 13: tb_getcolumn\n");
	time(&tt1);
	for(i = 0; i < 3; i++)
	{
	    struct tb_column col;
	    
	    tb_getcolumn(&t, i, &col);
	    printf("%s: %d %d\n", col.name, col.type, col.size); 
	}
	time(&tt2);
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");
	
	printf("test 14: sorting\n");
	time(&tt1);
	strcpy(col[0].name, "INT");
	col[0].type = tb_type_int;
	strcpy(col[1].name, "DBL");
	col[1].type = tb_type_double;
	strcpy(col[2].name, "CHR");
	col[2].type = tb_type_char;
	col[2].size = 128;
	strcpy(col[3].name, "RND");
	col[3].type = tb_type_int;
	tb_describe(&t, col, 4, 20000);
	tb_describe(&t1, col, 4, 20000);
	for(i = 0; i < TEST_ROWS; i++)
	{
	    char buf[20];
	    
	    tb_append(&t);
	    tb_setint(&t, i, 0, i);
	    tb_setdouble(&t, i, 1, i);
	    sprintf(buf, "%x", i);
	    tb_setchar(&t, i, 2, buf);
	    tb_setint(&t, i, 3, rand());
	    
	    tb_append(&t1);
	    tb_setint(&t1, i, 0, i);
	    tb_setdouble(&t1, i, 1, i);
	    tb_setchar(&t1, i, 2, buf);
	    tb_setint(&t1, i, 3, rand());
	}
	{
	    int j[1], o[1];

	    printf("sorting randomly ...\n");	    
	    j[0] = 3;
	    o[0] = 1;
	    tb_sort(&t, j, 1, o);
	    tb_sort(&t1, j, 1, o);
	}
	{
	    int j[1], o[1];
	    
	    j[0] = 2;
	    o[0] = 1;
	    tb_sort(&t, j, 1, o);
	    o[0] = -1;
	    tb_sort(&t1, j, 1, o);

	    printf("checking ...\n");	    
	    for(i = 0; i < tb_getnrows(&t); i++)
	    {
		char buf[20], buf1[20];
		
		tb_getchar(&t, i, 2, buf, sizeof(buf));
		tb_getchar(&t1, tb_getnrows(&t1) - i - 1, 2, buf1, sizeof(buf1));
		
		if(strcmp(buf, buf1))
		{
		    printf("sorting failure (%d): \"%s\"!=\"%s\"\n", i, buf, buf1);
		    break;
		}
	    }
	}
	time(&tt2);
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");
	
	printf("test 15: searching\n");
	time(&tt1);
	{
	    int j[1], o[1];
	    
	    printf("randomly sorting...\n");
	    j[0] = 3;
	    o[0] = 1;
	    tb_sort(&t, j, 1, o);
	    tb_sort(&t1, j, 1, o);
	}
	{
	    int j[2], o[2];
	    void *data[2];
	    int iv;
	    double dv;
	    char buf[20];
	    
	    printf("joining ...\n");
	    j[0] = 0;
	    j[1] = 1;
	    o[0] = 1;
	    o[1] = 1;
	    tb_sort(&t1, j, 2, o);
	    
	    data[0] = &iv;
	    data[1] = &dv;
	    
	    for(i = 0; i < tb_getnrows(&t); i++)
	    {
		int l;
		int itv;
		double dtv;
		char buf1[20];
		
		iv = tb_getint(&t, i, 0);
		dv = tb_getdouble(&t, i, 1);
		tb_getchar(&t, i, 2, buf, sizeof(buf));

		l = tb_bsearch(&t1, j, 2, data);
		if(l < 0)
		{
		    printf("join failure: %d %f (row %d)\n", iv, dv, i);
		    break;
		} 
		
		itv = tb_getint(&t1, l, 0);
		dtv = tb_getdouble(&t1, l, 1);
		tb_getchar(&t1, l, 2, buf1, sizeof(buf1));
		
		if(dv != dtv || iv != itv || strcmp(buf, buf1))
		{
		    printf("join failure (2): %f!=%f || %d!=%d\n || %s!=%s", dv, dtv, iv, itv, buf, buf1);
		    break;
		}
	    }
	}
	time(&tt2);
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");
	
	printf("test 16: tb_discard\n");
	time(&tt1);
	tb_discard(&t);
	tb_discard(&t1);
	time(&tt2);
	printf("%f sec\n", difftime(tt2, tt1));
	tott += difftime(tt2, tt1);
	printf("\n");
	
	printf("total time %f\n", tott);
	printf("\n");

  printf("total time: %f\n",((float)GetTickCount()-dw)/1000);

	printf("press enter ...\n");
	getchar();
	return 0;
}
